#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"
#include "xmlparse.h"

#define BUFSIZE 32768

#define XMLP_UPD(fld) \
  if (cbv->fld) {\
    if (cbv->fld != fld)\
      Perl_sv_setsv(cbv->fld, fld);\
  }\
  else\
    cbv->fld=newSVsv(fld)

typedef struct {
    SV* self_sv;
    char *buffstrt;
    int bufflen;
    int	offset;
    int prev_offset;
    SV* start_sv;
    SV* end_sv;
    SV* char_sv;
    SV* proc_sv;
    SV* dflt_sv;
    SV* unprsd_sv;
    SV* notation_sv;
    SV* extent_sv;
} CallbackVector;

static void
append_error(XML_Parser parser, char * err)
{
  dSP;
  CallbackVector * cbv;
  SV ** errstr;

  cbv = (CallbackVector*) XML_GetUserData(parser);
  errstr = hv_fetch((HV*)SvRV(cbv->self_sv),
		    "ErrorMessage", 12, 0);

  if (errstr && SvPOK(*errstr)) {
    SV ** errctx = hv_fetch((HV*) SvRV(cbv->self_sv),
			    "ErrorContext", 12, 0);
    int dopos = !err && errctx && SvOK(*errctx);

    if (! err)
      err = (char *) XML_ErrorString(XML_GetErrorCode(parser));

    sv_catpvf(*errstr, "\n%s at line %d, column %d, byte %d%s",
	      err,
	      XML_GetCurrentLineNumber(parser),
	      XML_GetCurrentColumnNumber(parser),
	      XML_GetCurrentByteIndex(parser),
	      dopos ? ":\n" : "");

    if (dopos)
      {
	int count;

	ENTER ;
	SAVETMPS ;
	PUSHMARK(sp);
	XPUSHs(cbv->self_sv);
	XPUSHs(*errctx);
	PUTBACK ;

	count = perl_call_method("position_in_context", G_SCALAR);

	SPAGAIN ;

	if (count >= 1) {
	  sv_catsv(*errstr, POPs);
	}

	PUTBACK ;
	FREETMPS ;
	LEAVE ;
      }
  }
}

static int
parse_stream(XML_Parser parser, SV * ioref, int close_it)
{
  PerlIO *fp;
  int fd;
  int br = 0;
  int ret = 1;
  CallbackVector *cbv;

  if (SvTYPE(ioref) != SVt_PVIO
      || ! (fp = IoIFP(ioref))
      || (fd = PerlIO_fileno(fp)) < 0)
    croak("Not an open IO ref");

  cbv = (CallbackVector*) XML_GetUserData(parser);

  for (cbv->offset = 0, cbv->prev_offset = 0;;
       cbv->prev_offset = cbv->offset, cbv->offset += br)
    {
      char *buff = XML_GetBuffer(parser, BUFSIZE);
      if (! buff)
	{
	  append_error(parser, "Ran out of memory");
	  return 0;
	}

      br = read(fd, buff, BUFSIZE);
      if (br < 0)
	{
	  append_error(parser, "Read error");
	  return 0;
	}

      if (br == 0)
	{
	  cbv->offset = cbv->prev_offset;
	}
      else
	{
	  cbv->buffstrt = buff;
	  cbv->bufflen = br;
	}

      ret = XML_ParseBuffer(parser, br, br == 0);

      if (!ret || br == 0)
	break;
    }

  if (! ret)
    append_error(parser, NULL);

  if (close_it)
    close(fd);

  return ret;
}

static void characterData(void *userData, const char *s, int len)
{
  dSP;
  CallbackVector* cbv = (CallbackVector*) userData;

  PUSHMARK(sp);
  XPUSHs(cbv->self_sv);
  XPUSHs(sv_2mortal(newSVpv((char*)s,len)));
  PUTBACK;
  perl_call_sv(cbv->char_sv, G_DISCARD);
}

static void startElement(void *userData, const char *name, const char **atts)
{
  dSP;
  CallbackVector* cbv = (CallbackVector*) userData;
  SV ** pcontext;

  if (SvTRUE(cbv->start_sv))
    {
      PUSHMARK(sp);
      XPUSHs(cbv->self_sv);
      XPUSHs(sv_2mortal(newSVpv((char*)name,0)));
      while (*atts)
	XPUSHs(sv_2mortal(newSVpv((char*)*atts++,0)));
      PUTBACK;
      perl_call_sv(cbv->start_sv, G_DISCARD);
    }

  pcontext = hv_fetch((HV*)SvRV(cbv->self_sv), "Context", 7, 0);
  if (pcontext && SvROK(*pcontext) && SvTYPE(SvRV(*pcontext)) == SVt_PVAV)
    {
      av_push((AV*)SvRV(*pcontext), newSVpv((char*)name, 0));
    }
}

static void endElement(void *userData, const char *name)
{
  dSP;
  CallbackVector* cbv = (CallbackVector*) userData;
  SV ** pcontext = hv_fetch((HV*)SvRV(cbv->self_sv), "Context", 7, 0);

  if (pcontext && SvROK(*pcontext) && SvTYPE(SvRV(*pcontext)) == SVt_PVAV)
    {
      SV * pv = av_pop((AV*) SvRV(*pcontext));
      SvREFCNT_dec(pv);
    }

  if (SvTRUE(cbv->end_sv))
    {
      PUSHMARK(sp);
      XPUSHs(cbv->self_sv);
      XPUSHs(sv_2mortal(newSVpv((char*)name,0)));
      PUTBACK;
      perl_call_sv(cbv->end_sv, G_DISCARD);
    }
}

static void processingInstruction(void *userData, const char *target, const char *data)
{
  dSP;
  CallbackVector* cbv = (CallbackVector*) userData;

  PUSHMARK(sp);
  XPUSHs(cbv->self_sv);
  XPUSHs(sv_2mortal(newSVpv((char*)target,0)));
  XPUSHs(sv_2mortal(newSVpv((char*)data,0)));
  PUTBACK;
  perl_call_sv(cbv->proc_sv, G_DISCARD);
}

static void defaulthandle(void *userData, const char *string, int len)
{
  dSP;
  CallbackVector* cbv = (CallbackVector*) userData;

  PUSHMARK(SP);
  XPUSHs(cbv->self_sv);
  XPUSHs(sv_2mortal(newSVpv((char*)string, len)));
  PUTBACK;
  perl_call_sv(cbv->dflt_sv, G_DISCARD);
}

static void unparsedEntityDecl(void *userData,
			       const char* entity,
			       const char* base,
			       const char* sysid,
			       const char* pubid,
			       const char* notation)
{
  dSP;
  CallbackVector* cbv = (CallbackVector*) userData;

  PUSHMARK(sp);
  XPUSHs(cbv->self_sv);
  XPUSHs(sv_2mortal(newSVpv((char*) entity, 0)));
  XPUSHs(base ? sv_2mortal(newSVpv((char*) base, 0)) : &sv_undef);
  XPUSHs(sv_2mortal(newSVpv((char*) sysid, 0)));
  XPUSHs(pubid ? sv_2mortal(newSVpv((char*) pubid, 0)) : &sv_undef);
  XPUSHs(sv_2mortal(newSVpv((char*) notation, 0)));
  PUTBACK;
  perl_call_sv(cbv->unprsd_sv, G_DISCARD);
}

static void notationDecl(void *userData,
			 const char *name,
			 const char *base,
			 const char *sysid,
			 const char *pubid)
{
  dSP;
  CallbackVector* cbv = (CallbackVector*) userData;

  PUSHMARK(sp);
  XPUSHs(cbv->self_sv);
  XPUSHs(sv_2mortal(newSVpv((char*) name, 0)));
  if (base)
    {
      XPUSHs(sv_2mortal(newSVpv((char *) base, 0)));
    }
  else if (sysid || pubid)
    {
      XPUSHs(&sv_undef);
    }

  if (sysid)
    {
      XPUSHs(sv_2mortal(newSVpv((char *) sysid, 0)));
    }
  else if (pubid)
    {
      XPUSHs(&sv_undef);
    }
  
  if (pubid)
    XPUSHs(sv_2mortal(newSVpv((char *) pubid, 0)));

  PUTBACK;
  perl_call_sv(cbv->notation_sv, G_DISCARD);
}

static int externalEntityRef(XML_Parser parser,
			     const char* open,
			     const char* base,
			     const char* sysid,
			     const char* pubid)
{
  dSP;

  int count;
  int ret = 0;
  int parse_done = 0;

  CallbackVector* cbv = (CallbackVector*) XML_GetUserData(parser);

  ENTER ;
  SAVETMPS ;
  PUSHMARK(sp);
  XPUSHs(cbv->self_sv);
  XPUSHs(base ? sv_2mortal(newSVpv((char*) base, 0)) : &sv_undef);
  XPUSHs(sv_2mortal(newSVpv((char*) sysid, 0)));
  if (pubid)
    XPUSHs(sv_2mortal(newSVpv((char*) pubid, 0)));
  PUTBACK ;
  count = perl_call_sv(cbv->extent_sv, G_SCALAR);

  SPAGAIN ;

  if (count >= 1)
    {
       SV * result = POPs;
       int type;

       if (result && (type = SvTYPE(result)) > 0)
         {
	   char *oldbuff;
	   int oldoff, oldlen;
	   XML_Parser entpar;
	   SV **pval = hv_fetch((HV*) SvRV(cbv->self_sv),
				  "Parser", 6, 0);

	   if (! pval || ! SvIOK(*pval))
	     croak("Can't get parser field");

	   entpar = XML_ExternalEntityParserCreate(parser, open, 0);

	   sv_setiv(*pval, (IV) entpar);

	   if (type == SVt_RV)
	     {
	       int rtype = SvTYPE(SvRV(result));

	       if (rtype == SVt_PVGV || rtype == SVt_PVIO)
		 {
		   type = rtype;
		   result = SvRV(result);
		 }
	     }

	   oldbuff = cbv->buffstrt;
	   oldoff = cbv->offset;
	   oldlen = cbv->bufflen;

	   switch (type)
	     {
	     case SVt_PVGV:
		result = (SV *) GvIO(result);
		/* Fall through */

	      case SVt_PVIO:
                ret = parse_stream(entpar, result, 1);
		break;

              default:
                if (SvPOK(result))
                {
		  STRLEN  eslen;
                  int pret;
		  char *entstr = SvPV(result, eslen);

		  cbv->buffstrt = entstr;
		  cbv->offset   = 0;
		  cbv->bufflen  = eslen;
		  ret = XML_Parse(entpar, entstr, eslen, 1);

		  if (! ret)
		    append_error(entpar, NULL);
                }
              }

	    parse_done = 1;
	    cbv->buffstrt = oldbuff;
	    cbv->offset = oldoff;
	    cbv->bufflen = oldlen;

	    sv_setiv(*pval, (IV) parser);

	    XML_ParserFree(entpar);
         }
    }

  if (! ret && ! parse_done)
    append_error(parser, "Handler couldn't resolve external entity");

  PUTBACK ;
  FREETMPS ;
  LEAVE ;

  return ret;
}

MODULE = XML::Parser	PACKAGE = XML::Parser	PREFIX = XML_

XML_Parser
XML_ParserCreate(enc_sv, self_sv)
	SV *			enc_sv
        SV *                    self_sv
    CODE:
	{
	    CallbackVector *cbv = malloc(sizeof(CallbackVector));
	    char *enc = (char *) (SvTRUE(enc_sv) ? SvPV(enc_sv,na) : 0);
	    memset(cbv, 0, sizeof(CallbackVector));
	    cbv->self_sv = SvREFCNT_inc(self_sv);	/* XXX makes loop */
	    RETVAL =  XML_ParserCreate(enc);
	    XML_SetUserData(RETVAL, (void *) cbv);
	    XML_SetElementHandler(RETVAL, startElement, endElement);
	}
    OUTPUT:
	RETVAL

void
XML_ParserFree(parser)
	XML_Parser parser
    CODE:
	free(XML_GetUserData(parser));
	XML_ParserFree(parser);

int
XML_ParseString(parser, s)
        XML_Parser			parser
	char *				s
	int				len = na;
    CODE:
        {
	  CallbackVector * cbv;

          cbv = (CallbackVector *) XML_GetUserData(parser);

	  cbv->buffstrt = s;
	  cbv->offset = 0;
	  cbv->bufflen = len;
	  RETVAL = XML_Parse(parser, s, len, 1);
	  if (! RETVAL)
	    append_error(parser, NULL);
	}

    OUTPUT:
	RETVAL

int
XML_ParseStream(parser, ioref)
	XML_Parser			parser
	SV *				ioref
    CODE:
        if (!SvROK(ioref))
          croak("Argument is not a ref");

	RETVAL = parse_stream(parser, SvRV(ioref), 0);

    OUTPUT:
	RETVAL

void
XML_SetStartElementHandler(parser, start_sv)
	XML_Parser			parser
	SV *				start_sv
    CODE:
	{
	  CallbackVector * cbv = (CallbackVector*) XML_GetUserData(parser);
	  XMLP_UPD(start_sv);
	}

void
XML_SetEndElementHandler(parser, end_sv)
	XML_Parser			parser
	SV *				end_sv
    CODE:
	{
	  CallbackVector *cbv = (CallbackVector*) XML_GetUserData(parser);
	  XMLP_UPD(end_sv);
	}

void
XML_SetCharacterDataHandler(parser, char_sv)
	XML_Parser			parser
	SV *				char_sv
    CODE:
	{
	  XML_CharacterDataHandler charhndl = (XML_CharacterDataHandler) 0;
	  CallbackVector * cbv = (CallbackVector*) XML_GetUserData(parser);

	  if (SvTRUE(char_sv))
	    {
	      XMLP_UPD(char_sv);
	      charhndl = characterData;
	    }

	  XML_SetCharacterDataHandler(parser, charhndl);
	}

void
XML_SetProcessingInstructionHandler(parser, proc_sv)
	XML_Parser			parser
	SV *				proc_sv
    CODE:
	{
	  XML_ProcessingInstructionHandler prochndl =
	    (XML_ProcessingInstructionHandler) 0;
	  CallbackVector* cbv = (CallbackVector*) XML_GetUserData(parser);

	  if (SvTRUE(proc_sv))
	    {
	      XMLP_UPD(proc_sv);
	      prochndl = processingInstruction;
	    }

	  XML_SetProcessingInstructionHandler(parser, prochndl);
	}

void
XML_SetDefaultHandler(parser, dflt_sv)
	XML_Parser			parser
	SV *				dflt_sv
    CODE:
	{
	  XML_DefaultHandler dflthndl = (XML_DefaultHandler) 0;
	  CallbackVector * cbv = (CallbackVector*) XML_GetUserData(parser);

	  if (SvTRUE(dflt_sv))
	    {
	      XMLP_UPD(dflt_sv);
	      dflthndl = defaulthandle;
	    }
	  else
	    cbv->dflt_sv = &sv_undef;

	  XML_SetDefaultHandler(parser, dflthndl);
	}

void
XML_SetUnparsedEntityDeclHandler(parser, unprsd_sv)
	XML_Parser			parser
	SV *				unprsd_sv
    CODE:
	{
	  XML_UnparsedEntityDeclHandler unprsdhndl =
	    (XML_UnparsedEntityDeclHandler) 0;
	  CallbackVector * cbv = (CallbackVector*) XML_GetUserData(parser);

	  if (SvTRUE(unprsd_sv))
	    {
	      XMLP_UPD(unprsd_sv);
	      unprsdhndl = unparsedEntityDecl;
	    }

	  XML_SetUnparsedEntityDeclHandler(parser, unprsdhndl);
	}

void
XML_SetNotationDeclHandler(parser, notation_sv)
	XML_Parser			parser
	SV *				notation_sv
    CODE:
	{
	  XML_NotationDeclHandler nothndlr = (XML_NotationDeclHandler) 0;
	  CallbackVector * cbv = (CallbackVector*) XML_GetUserData(parser);

	  if (SvTRUE(notation_sv))
	    {
	      XMLP_UPD(notation_sv);
	      nothndlr = notationDecl;
	    }

	  XML_SetNotationDeclHandler(parser, nothndlr);
	}

void
XML_SetExternalEntityRefHandler(parser, extent_sv)
	XML_Parser			parser
	SV *				extent_sv
    CODE:
	{
	  XML_ExternalEntityRefHandler exthndlr =
	    (XML_ExternalEntityRefHandler) 0;
	  CallbackVector * cbv = (CallbackVector*) XML_GetUserData(parser);

	  if (SvTRUE(extent_sv))
	    {
	      XMLP_UPD(extent_sv);
	      exthndlr = externalEntityRef;
	    }

	  XML_SetExternalEntityRefHandler(parser, exthndlr);
	}
	   
int
XML_SetBase(parser, base)
	XML_Parser			parser
	char*				base

char*
XML_GetBase(parser)
	XML_Parser			parser
    CODE:
	const char *ret = XML_GetBase(parser);
	ST(0) = sv_newmortal();
	sv_setpv((SV*)ST(0), ret);

void
XML_PositionContext(parser, lines)
	XML_Parser			parser
	int				lines
    PREINIT:
	CallbackVector *cbv = (CallbackVector *) XML_GetUserData(parser);
        char *pos = cbv->buffstrt;
	char *markbeg, *markend, *limit;
	int length, relpos;
	int  cnt;
	int parsepos = XML_GetCurrentByteIndex(parser);

    PPCODE:
          parsepos -= cbv->offset;
	  if (parsepos < 0)
	    parsepos = 0;

	  if (parsepos >= cbv->bufflen)
	    croak("Parse position is outside of buffer");

	  for (markbeg = &pos[parsepos], cnt = 0; markbeg >= pos; markbeg--)
	    {
	      if (*markbeg == '\n')
		{
		  cnt++;
		  if (cnt > lines)
		    break;
		}
	    }

	  markbeg++;

          relpos = 0;
	  limit = &pos[cbv->bufflen];
	  for (markend = &pos[parsepos + 1], cnt = 0;
	       markend < limit;
	       markend++)
	    {
	      if (*markend == '\n')
		{
		  if (cnt == 0)
                     relpos = (markend - markbeg) + 1;
		  cnt++;
		  if (cnt > lines)
		    {
		      markend++;
		      break;
		    }
		}
	    }

	  length = markend - markbeg;
          if (relpos == 0)
            relpos = length;

          EXTEND(sp, 2);
	  PUSHs(sv_2mortal(newSVpv(markbeg, length)));
	  PUSHs(sv_2mortal(newSViv(relpos)));

void
XML_DefaultCurrent(parser)
	XML_Parser			parser

int
XML_GetErrorCode(parser)
	XML_Parser			parser

int
XML_GetCurrentLineNumber(parser)
	XML_Parser			parser


int
XML_GetCurrentColumnNumber(parser)
	XML_Parser			parser

long
XML_GetCurrentByteIndex(parser)
	XML_Parser			parser

char *
XML_ErrorString(code)
	int				code
    CODE:
	const char *ret = XML_ErrorString(code);
	ST(0) = sv_newmortal();
	sv_setpv((SV*)ST(0), ret);

